from .abstract import AbstractLLM
from langchain.prompts import ChatPromptTemplate
from langchain.memory import ConversationBufferWindowMemory

from langchain.chains.llm import LLMChain
from langchain_openai.chat_models.base import ChatOpenAI


class DialogOpenAI(AbstractLLM):
    def __init__(self, *args, **kwargs):
        model = kwargs.pop("model", "gpt-3.5-turbo")
        temperature = kwargs.pop("temperature", 0.1)
        kwargs["config"] = kwargs.get("config", {})

        self.memory_instance = kwargs.pop("memory", None)
        self.llm_api_key = kwargs.pop("openai_api_key", None)
        self.prompt_content = kwargs.pop("prompt", None)

        super().__init__(*args, **kwargs)

        self.chat_model = ChatOpenAI(
            openai_api_key=self.llm_api_key,
            model=model,
            temperature=temperature
        )

    def generate_prompt(self, input_text):
        self.prompt = ChatPromptTemplate.from_messages([
            ("ai", self.prompt_content),
            ("human", input_text)
        ])
        return input_text

    @property
    def memory(self):
        return self.memory_instance

    @property
    def llm(self):
        chain_settings = dict(
            llm=self.chat_model,
            prompt=self.prompt
        )

        if self.memory:
            buffer_config = {
                "chat_memory": self.memory,
                "memory_key": "chat_history",
                "return_messages": True,
                "k": self.config.get("memory_size", 5)
            }
            chain_settings["memory"] = ConversationBufferWindowMemory(
                **buffer_config
            )

        return LLMChain(
            **chain_settings
        )

    def postprocess(self, output):
        return output.get("text")


import pytest
from dialog_lib.agents.abstract import AbstractLLM


def test_abstract_agent_with_invalid_config():
    with pytest.raises(ValueError):
        AbstractLLM(config="invalid_config")


def test_abstract_agent_with_valid_config():
    config = {
        "model": "gpt3.5-turbo",
        "temperature": 0.5,
    }
    agent = AbstractLLM(config=config)
    assert agent.config == config
    assert agent.prompt is None
    assert agent.session_id is None
    assert agent.relevant_contents is None
    assert agent.dataset is None
    assert agent.llm_api_key is None
    assert agent.parent_session_id is None
    assert agent.dbsession is None

def test_abstract_agent_get_prompt():
    config = {
        "model": "gpt3.5-turbo",
        "temperature": 0.5,
    }
    agent = AbstractLLM(config=config)

def test_abstract_agent_memory():
    config = {
        "model": "gpt3.5-turbo",
        "temperature": 0.5,
    }
    agent = AbstractLLM(config=config)
    with pytest.raises(NotImplementedError):
        agent.memory

def test_pre_and_post_processing():
    config = {
        "model": "gpt3.5-turbo",
        "temperature": 0.5,
    }
    agent = AbstractLLM(config=config)
    output = agent.preprocess(input="Hello")
    assert output == "Hello"

    post_processed_output = agent.postprocess(output="Hello")
    assert post_processed_output == "Hello"

def test_abstract_agent_generate_prompt():
    config = {
        "model": "gpt3.5-turbo",
        "temperature": 0.5,
    }
    agent = AbstractLLM(config=config)
    prompt = agent.generate_prompt(text="Hello")
    assert prompt == "Hello"

def test_abstract_agent_process(mocker):
    config = {
        "model": "gpt3.5-turbo",
        "temperature": 0.5,
        "prompt": {
            "fallback_not_found_relevant_contents": "404 Not Found"
        }
    }
    agent = AbstractLLM(config=config)

    agent.relevant_contents = [] # sample mock
    mocker.patch('dialog_lib.agents.abstract.AbstractLLM.llm')
    mocker.patch('dialog_lib.agents.abstract.AbstractLLM.llm.invoke', return_value={'text': '404 Not Found'})
    output = agent.process(input="Hello")
    assert output == {'text': '404 Not Found'}



from langchain.prompts import ChatPromptTemplate
from langchain.chains.llm import LLMChain
from langchain.memory.chat_memory import BaseChatMemory


class AbstractLLM:
    def __init__(
        self,
        config,
        session_id=None,
        parent_session_id=None,
        dataset=None,
        llm_api_key=None,
        dbsession=None,
    ):
        """
        :param config: Configuration dictionary

        The constructor of the AbstractLLM class allows users to pass
        a configuration dictionary to the LLM. This configuration dictionary
        can be used to configure the LLM temperature, prompt and other
        necessities.
        """
        if not isinstance(config, dict):
            raise ValueError("Config must be a dictionary")

        self.config = config
        self.prompt = None
        self.session_id = None
        self.relevant_contents = None
        if session_id:
            self.session_id = (
                session_id if dataset is None else f"{dataset}_{session_id}"
            )
        self.dataset = dataset
        self.llm_api_key = self.config.get("llm_api_key", llm_api_key)
        self.parent_session_id = parent_session_id
        self.dbsession = dbsession

    @property
    def memory(self) -> BaseChatMemory:
        """
        Returns the memory instance
        """
        raise NotImplementedError("Memory must be implemented")

    @property
    def llm(self) -> LLMChain:
        """
        Returns the LLM instance. If a memory instance is provided,
        the LLM instance should be initialized with the memory instance.

        If no memory instance is provided, the LLM instance should be
        initialized without a memory instance.
        """
        raise NotImplementedError("LLM must be implemented")

    def preprocess(self, input: str) -> str:
        """
        Function that pre-process the LLM input, enabling users
        to modify the input before it is processed by the LLM.

        This can be used to add context or prefixes to the LLM.
        """
        return input

    def generate_prompt(self, text: str) -> str:
        """
        Function that generates the prompt using PromptTemplate for the LLM.
        """
        return text

    def postprocess(self, output: str) -> str:
        """
        Function that post-process the LLM output, enabling users
        to modify the output before it is returned to the user.
        """
        return output

    def process(self, input: str):
        """
        Function that encapsulates the pre-processing, processing and post-processing
        of the LLM.
        """
        processed_input = self.preprocess(input)
        self.generate_prompt(processed_input)
        output = self.llm.invoke(
            {
                "user_message": processed_input,
            }
        )
        processed_output = self.postprocess(output)
        return processed_output

    @property
    def messages(self):
        """
        Returns the messages from the memory instance
        """
        return self.memory.messages


class AbstractRAG(AbstractLLM):
    relevant_contents = []

    def process(self, input: str):
        """
        Function that encapsulates the pre-processing, processing and post-processing
        of the LLM.
        """
        processed_input = self.preprocess(input)
        self.generate_prompt(processed_input)
        if len(self.relevant_contents) == 0 and self.config.get("prompt").get(
            "fallback_not_found_relevant_contents"
        ):
            return {
                "text": self.config.get("prompt").get(
                    "fallback_not_found_relevant_contents"
                )
            }
        output = self.llm.invoke(
            {
                "user_message": processed_input,
            }
        )
        processed_output = self.postprocess(output)
        return processed_output


import os
import sqlalchemy

from dialog_lib.db import (
    generate_memory_instance, add_user_message_to_message_history, get_messages,
    Chat, ChatMessages, CompanyContent
)
from dialog_lib.db.memory import CustomPostgresChatMessageHistory

from langchain_core.messages import HumanMessage


def test_models_were_created(db_engine):
    assert True == sqlalchemy.inspect(db_engine).has_table(Chat.__tablename__)
    assert True == sqlalchemy.inspect(db_engine).has_table(ChatMessages.__tablename__)
    assert True == sqlalchemy.inspect(db_engine).has_table(CompanyContent.__tablename__)

def test_generate_memory_instance(db_session):
    memory = generate_memory_instance(
        "test_session_id",
        database_url=os.environ.get('DATABASE_URL'),
        dbsession=db_session,
    )
    assert isinstance(memory, CustomPostgresChatMessageHistory)

def test_add_user_message_to_memory_instance_and_db(db_session):
    memory = add_user_message_to_message_history(
        "test_session_id",
        "test_message",
        dbsession=db_session,
        database_url=os.environ.get('DATABASE_URL'),
    )
    assert isinstance(memory.messages[0], HumanMessage)
    assert memory.messages[0].content == "test_message"

def test_get_messages(db_session):
    add_user_message_to_message_history(
        "test_session_id",
        "test_message",
        dbsession=db_session,
        database_url=os.environ.get('DATABASE_URL'),
    )
    messages = get_messages(
        "test_session_id",
        dbsession=db_session,
        database_url=os.environ.get('DATABASE_URL'),
    )
    assert isinstance(messages[0], HumanMessage)
    assert messages[0].content == "test_message"